from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from dcqc.file import File, FileType


def test_for_an_error_if_registering_a_duplicate_file_type():
    with pytest.raises(ValueError):
        FileType("txt", (".foo",))


def test_for_an_error_when_requesting_for_an_unregistered_file_type():
    with pytest.raises(ValueError):
        FileType.get_file_type("foo")


def test_for_an_error_when_retrieving_missing_metadata_on_a_file(test_files):
    test_file = test_files["good"]
    with pytest.raises(KeyError):
        test_file.get_metadata("foo")


def test_that_a_local_file_is_not_moved_when_requesting_a_local_path(test_files):
    test_file = test_files["good"]
    url_before = test_file.url
    local_path = test_file.local_path
    url_after = test_file.url
    assert url_before == url_after
    assert local_path.exists()


def test_for_an_error_when_accessing_local_path_of_an_unstaged_remote_file(test_files):
    remote_file = test_files["remote"]
    with pytest.raises(FileNotFoundError):
        remote_file.local_path


def test_that_a_local_file_is_not_moved_when_staged_without_a_destination(test_files):
    test_file = test_files["good"]
    path_before = test_file.local_path
    path_after = test_file.stage()
    assert path_before == path_after


def test_that_a_local_file_is_symlinked_when_staged_with_a_destination(test_files):
    test_file = test_files["good"]
    with TemporaryDirectory() as tmp_dir:
        original_path = Path(test_file.local_path)
        tmp_dir_path = Path(tmp_dir)
        test_file.stage(tmp_dir_path)
        staged_path = Path(test_file.local_path)
        assert staged_path.is_symlink()
        assert staged_path.resolve() == original_path.resolve()


def test_that_a_local_temporary_path_is_created_when_staging_a_remote_file(test_files):
    remote_file = test_files["remote"]
    staged_path = remote_file.stage()
    assert staged_path.exists()
    assert remote_file.local_path == staged_path


def test_that_a_remote_file_is_created_when_staged_with_a_destination(test_files):
    remote_file = test_files["remote"]
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir) / "test.txt"
        staged_path = remote_file.stage(tmp_path)
        assert staged_path.exists()
        assert staged_path == tmp_path
        assert remote_file.local_path == staged_path


def test_that_a_file_can_be_saved_and_restored_without_changing(test_files):
    file_1 = test_files["good"]
    file_1_dict = file_1.to_dict()
    file_2 = File.from_dict(file_1_dict)
    file_2_dict = file_2.to_dict()
    assert file_1 == file_2
    assert file_1_dict == file_2_dict


def test_that_an_absolute_local_url_is_unchanged_when_using_relative_to(get_data):
    test_path = get_data("test.txt")
    test_url = test_path.resolve().as_posix()
    metadata = {"file_type": "TXT"}
    file = File(test_url, metadata, relative_to=Path.cwd())
    assert file.url == test_url


def test_that_an_fs_is_created_when_an_fs_path_is_requested(test_files):
    file = test_files["good"]
    assert file._fs_path is None
    file.fs_path
    assert file._fs_path is not None
    assert file._fs is not None


def test_that_file_name_is_cached(test_files):
    file = test_files["remote"]
    assert file._name is None
    attempt_1 = file.name
    assert file._name is not None
    attempt_2 = file.name
    assert attempt_1 is attempt_2


def test_for_an_error_when_staging_a_file_where_one_already_exists(test_files):
    remote_file = test_files["remote"]
    existing_file = test_files["good"]
    with pytest.raises(FileExistsError):
        remote_file.stage(existing_file.local_path)


def test_for_an_error_when_staging_a_file_under_a_nonexistent_directory(test_files):
    remote_file = test_files["remote"]
    invalid_destination = Path("./nonexistent_dir/test.txt")
    with pytest.raises(ValueError):
        remote_file.stage(invalid_destination)


def test_that_an_unset_local_path_is_ignored_during_deserialization(test_files):
    file = test_files["remote"]
    dictionary = file.to_dict()
    file_from_dict = File.from_dict(dictionary)
    with pytest.raises(FileNotFoundError):
        file_from_dict.local_path


def test_that_a_file_cannot_be_made_relative_to_a_nonexistent_directory(test_files):
    file = test_files["good"]
    nonexistent_dir = Path("foobar")
    with pytest.raises(ValueError):
        file.serialize_paths_relative_to(nonexistent_dir)


def test_that_a_file_cannot_be_made_relative_to_a_another_file(test_files):
    file = test_files["good"]
    another_file = test_files["bad"]
    another_file_path = another_file.local_path
    with pytest.raises(ValueError):
        file.serialize_paths_relative_to(another_file_path)


def test_that_paths_are_unchanged_when_not_using_serialize_paths_relative_to():
    path = Path("foobar")
    file = File("test.txt", local_path=path)
    file_dict = file.to_dict()
    assert file_dict["local_path"] == str(path)


def test_that_paths_change_when_using_serialize_paths_relative_to(get_output):
    path = Path("foo")
    relative_to = get_output("relative-to")
    relative_to.mkdir(parents=True, exist_ok=True)
    file = File("test.txt", local_path=path)
    file.serialize_paths_relative_to(relative_to)
    file_dict = file.to_dict()
    assert file_dict["local_path"] != str(path)



from __future__ import annotations

from abc import ABC
from collections.abc import Collection, Sequence
from copy import deepcopy
from enum import Enum
from typing import ClassVar, Generic, Optional, Type, TypeVar, Union

from dcqc.file import FileType
from dcqc.mixins import SerializableMixin, SerializedObject, SubclassRegistryMixin
from dcqc.target import BaseTarget, SingleTarget
from dcqc.tests import BaseTest, TestStatus

Target = TypeVar("Target", bound=BaseTarget)


class SuiteStatus(Enum):
    NONE = "NONE"  # status not yet evaluated
    GREEN = "GREEN"  # all tests passed
    RED = "RED"  # one or more required tests failed
    AMBER = "AMBER"  # all required tests passed, but one or more optional tests failed
    # TODO GREY = "GREY" # error occurred


# TODO: Consider the Composite design pattern once
#       we have higher-level QC suites
class SuiteABC(SerializableMixin, SubclassRegistryMixin, ABC, Generic[Target]):
    """Abstract base class for QC test suites.

    Args:
        target: Single- or multi-file target.
        required_tests: List of tests that must pass for the
            overall suite to pass. Defaults to None,
            which requires tier-1 and tier-2 tests.
        skipped_tests: List of tests that should not be
            evaluated. Defaults to None.
    """

    # Class attributes
    file_type: ClassVar[FileType]
    add_tests: ClassVar[tuple[Type[BaseTest], ...]]
    del_tests: ClassVar[tuple[Type[BaseTest], ...]]

    # Instance attributes
    type: str
    target: Target
    required_tests: set[str]
    skipped_tests: set[str]

    def __init__(
        self,
        target: Target,
        required_tests: Optional[Collection[str]] = None,
        skipped_tests: Optional[Collection[str]] = None,
    ):
        self.type = self.__class__.__name__
        self.target = target

        test_classes = self.list_test_classes()
        test_names = set(test.__name__ for test in test_classes)

        # To differentiate between None and []
        if required_tests is None:
            required_tests = self._default_required_tests()
        self.required_tests = set(required_tests).intersection(test_names)

        skipped_tests = skipped_tests or list()
        self.skipped_tests = set(skipped_tests).intersection(test_names)

        self.tests = self.init_test_classes()
        self._status = SuiteStatus.NONE

    @classmethod
    def from_target(
        cls,
        target: SingleTarget,
        required_tests: Optional[Collection[str]] = None,
        skipped_tests: Optional[Collection[str]] = None,
    ) -> SuiteABC:
        """Generate a suite from a single-file target.

        The suite is selected based on the target file type.

        Args:
            target: A QC target.
            required_tests: List of requires tests.
                Defaults to None, which requires tier-1
                and tier-2 tests.
            skipped_tests: List of skipped tests.
                Defaults to None.

        Returns:
            SuiteABC: An initialized test suite.
        """
        file_type = target.get_file_type()
        suite_cls = SuiteABC.get_subclass_by_file_type(file_type)
        suite = suite_cls(target, required_tests, skipped_tests)
        return suite

    @classmethod
    def from_tests(
        cls,
        tests: Sequence[BaseTest],
        required_tests: Optional[Collection[str]] = None,
        skipped_tests: Optional[Collection[str]] = None,
    ) -> SuiteABC:
        """Generate a suite from a set of tests.

        The tests must all have the same target.

        Args:
            tests: Set of tests with the same target.
            required_tests: List of requires tests.
                Defaults to None, which requires tier-1
                and tier-2 tests.
            skipped_tests: List of skipped tests.
                Defaults to None.

        Returns:
            SuiteABC: An initialized test suite.
        """
        targets = list()
        suite_tests = list()
        skipped_tests = skipped_tests or list()
        skipped_tests = set(skipped_tests)
        for test in tests:
            test_copy = test.copy()
            if test_copy.type in skipped_tests:
                test_copy.skip()
            targets.append(test_copy.target)
            suite_tests.append(test_copy)

        representative_target = targets[0]
        if not all(representative_target == target for target in targets):
            message = f"Not all tests refer to the same target ({targets})."
            raise ValueError(message)
        if not isinstance(representative_target, SingleTarget):  # pragma: no cover
            raise ValueError("Can only recreate suite from single-file target.")
        suite = cls.from_target(representative_target, required_tests, skipped_tests)
        suite.tests = suite_tests

        return suite

    @classmethod
    def list_test_classes(cls) -> tuple[Type[BaseTest], ...]:
        """List all applicable test classes"""
        all_tests: set[Type[BaseTest]]
        all_tests = set()

        superclasses = cls.__mro__
        for cls in reversed(superclasses):  # Start from the base class
            if hasattr(cls, "add_tests"):
                add_tests = set(cls.add_tests)  # type: ignore
                all_tests.update(add_tests)
            if hasattr(cls, "del_tests"):
                del_tests = set(cls.del_tests)  # type: ignore
                all_tests.difference_update(del_tests)

        return tuple(all_tests)

    @classmethod
    def list_test_classes_by_file_type(cls) -> dict[str, list[Type[BaseTest]]]:
        """List test classes by file type."""
        result = dict()
        suite_classes = cls.list_subclasses()
        for suite_cls in suite_classes:
            file_type = suite_cls.file_type.name
            test_classes = suite_cls.list_test_classes()
            result[file_type] = list(test_classes)
        return result

    @classmethod
    def _default_required_tests(cls) -> list[str]:
        test_classes = cls.list_test_classes()
        required_tests = filter(lambda test: test.tier <= 2, test_classes)
        required_test_names = [test.__name__ for test in required_tests]
        return required_test_names

    def init_test_classes(self) -> list[BaseTest]:
        """Initialize applicable test classes with target."""
        test_classes = self.list_test_classes()
        tests = []
        for test_cls in test_classes:
            test_name = test_cls.__name__
            skip = test_name in self.skipped_tests
            test = test_cls(self.target, skip)
            tests.append(test)
        return tests

    @classmethod
    def get_subclass_by_file_type(
        cls, file_type: Union[str, FileType]
    ) -> Type[SuiteABC]:
        """Retrieve a subclass by file type."""
        if isinstance(file_type, str):
            try:
                file_type = FileType.get_file_type(file_type)
            except ValueError:
                file_type = FileType.get_file_type("*")
        name = file_type.name
        subclasses = cls.list_subclasses()
        registry = {subcls.file_type.name: subcls for subcls in subclasses}
        if name not in registry:
            # TODO: This might have to be changed if we introduce
            #       composite file types (e.g., BAM/BAI file pair)
            return registry["*"]
        return registry[name]

    @property
    def tests_by_name(self):
        return {test.type: test for test in self.tests}

    def compute_tests(self) -> None:
        """Compute the status for each initialized test."""
        self.target.stage()
        for test in self.tests:
            test.get_status()

    def compute_status(self) -> SuiteStatus:
        """Compute the overall suite status."""
        self.compute_tests()
        if self._status != SuiteStatus.NONE:
            return self._status
        self._status = SuiteStatus.GREEN
        for test in self.tests:
            test_name = test.type
            test_status = test.get_status()
            if test_name in self.required_tests:
                if test_status == TestStatus.FAIL:
                    self._status = SuiteStatus.RED
                    return self._status
            else:
                if test_status == TestStatus.FAIL:
                    self._status = SuiteStatus.AMBER
        return self._status

    def to_dict(self) -> SerializedObject:
        suite_status = self.compute_status()
        test_dicts = []
        for test in self.tests:
            test_dict = test.to_dict()
            test_dict.pop("target", None)  # Remove redundant `target` info
            test_dicts.append(test_dict)
        suite_dict = {
            "type": self.type,
            "target": self.target.to_dict(),
            "suite_status": {
                "required_tests": list(self.required_tests),
                "skipped_tests": list(self.skipped_tests),
                "status": suite_status.value,
            },
            "tests": test_dicts,
        }
        return suite_dict

    @classmethod
    def from_dict(cls, dictionary: SerializedObject) -> SuiteABC:
        """Deserialize a dictionary into a suite.

        Args:
            dictionary: A serialized suite object.

        Returns:
            The reconstructed suite object.
        """
        dictionary = deepcopy(dictionary)

        suite_cls_name = dictionary["type"]
        suite_cls = SuiteABC.get_subclass_by_name(suite_cls_name)

        target_dict = dictionary["target"]
        target = BaseTarget.from_dict(target_dict)

        required_tests = dictionary["suite_status"]["required_tests"]
        skipped_tests = dictionary["suite_status"]["skipped_tests"]
        suite = suite_cls(target, required_tests, skipped_tests)

        suite_status = SuiteStatus(dictionary["suite_status"]["status"])
        suite._status = suite_status

        tests = list()
        for test_dict in dictionary["tests"]:
            test_dict["target"] = target_dict
            test = BaseTest.from_dict(test_dict)
            tests.append(test)
        suite.tests = tests

        return suite

    @classmethod
    def get_base_class(cls):
        """Retrieve base class."""
        return SuiteABC

    def get_status(self) -> SuiteStatus:
        """Compute (if applicable) and return the suite status."""
        if self._status == SuiteStatus.NONE:
            self._status = self.compute_status()
        return self._status



import json
from pathlib import Path

from dcqc.target import SingleTarget
from dcqc.tests.base_test import InternalBaseTest, TestStatus


class JsonLoadTest(InternalBaseTest):
    tier = 2
    target: SingleTarget

    def compute_status(self) -> TestStatus:
        path = self.target.file.stage()
        if self._can_be_loaded(path):
            status = TestStatus.PASS
        else:
            status = TestStatus.FAIL
        return status

    def _can_be_loaded(self, path: Path) -> bool:
        success = True
        with path.open("r") as infile:
            try:
                json.load(infile)
            except Exception:
                success = False
        return success



from unittest.mock import patch

import pytest

from dcqc.file import FileType
from dcqc.suites.suite_abc import SuiteABC, SuiteStatus
from dcqc.suites.suites import FileSuite, OmeTiffSuite, TiffSuite
from dcqc.tests import (
    BaseTest,
    FileExtensionTest,
    GrepDateTest,
    LibTiffInfoTest,
    TestStatus,
    TiffTag306DateTimeTest,
)

FileType("None", ())
FileType("Unpaired", ())


class RedundantFileSuite(TiffSuite):
    file_type = FileType.get_file_type("None")
    del_tests = (LibTiffInfoTest, GrepDateTest, TiffTag306DateTimeTest)


class DummyTest(BaseTest):
    def compute_status(self) -> TestStatus:
        return TestStatus.NONE


def test_that_a_file_suite_results_in_multiple_tests():
    tests = FileSuite.list_test_classes()
    assert len(tests) > 0
    assert all(issubclass(test, BaseTest) for test in tests)


def test_that_deleting_a_just_added_test_results_in_the_same_test_list():
    tests_1 = FileSuite.list_test_classes()
    tests_2 = RedundantFileSuite.list_test_classes()
    assert set(tests_1) == set(tests_2)


def test_that_the_ome_tiff_suite_has_a_superset_of_the_tiff_suite_tests():
    tiff_tests = TiffSuite.list_test_classes()
    ome_tiff_tests = OmeTiffSuite.list_test_classes()
    assert set(ome_tiff_tests) > set(tiff_tests)


def test_that_a_test_suite_can_be_retrieved_by_name():
    actual = SuiteABC.get_subclass_by_name("OmeTiffSuite")
    assert actual is OmeTiffSuite


def test_for_an_error_when_retrieving_a_nonexistent_test_suite_by_name():
    with pytest.raises(ValueError):
        SuiteABC.get_subclass_by_name("FooBarSuite")


def test_that_a_test_suite_can_be_retrieved_by_file_type_class():
    file_type = FileType.get_file_type("OME-TIFF")
    actual = SuiteABC.get_subclass_by_file_type(file_type)
    assert actual is OmeTiffSuite


def test_that_a_test_suite_can_be_retrieved_by_file_type_str():
    actual = SuiteABC.get_subclass_by_file_type("OME-TIFF")
    assert actual is OmeTiffSuite


def test_that_the_generic_file_suite_is_retrieved_for_a_random_file_type():
    actual = SuiteABC.get_subclass_by_file_type("Foo-Bar")
    assert actual is FileSuite


def test_that_the_generic_file_suite_is_retrieved_for_an_unpaired_file_type():
    actual = SuiteABC.get_subclass_by_file_type("Unpaired")
    assert actual is FileSuite


def test_that_the_default_required_tests_are_only_tiers_1_and_2(test_suites):
    suite = test_suites["jsonld"]
    assert all(test.tier <= 2 for test in suite.tests)


def test_that_skipped_tests_are_skipped_when_building_suite_from_tests(test_suites):
    suite = test_suites["tiff"]
    tests = suite.tests
    new_suite = SuiteABC.from_tests(tests, skipped_tests=["LibTiffInfoTest"])
    skipped_test_before = suite.tests_by_name["LibTiffInfoTest"]
    skipped_test_after = new_suite.tests_by_name["LibTiffInfoTest"]
    assert skipped_test_before.get_status(compute_ok=False) != TestStatus.SKIP
    assert skipped_test_after.get_status(compute_ok=False) == TestStatus.SKIP


def test_for_an_error_when_building_suite_from_tests_with_diff_targets(test_targets):
    target_1 = test_targets["good"]
    target_2 = test_targets["bad"]
    test_1 = FileExtensionTest(target_1)
    test_2 = FileExtensionTest(target_2)
    tests = [test_1, test_2]
    with pytest.raises(ValueError):
        SuiteABC.from_tests(tests)


def test_that_a_suite_will_consider_non_required_failed_tests(test_targets):
    target = test_targets["bad"]
    required_tests = []
    skipped_tests = ["LibTiffInfoTest", "GrepDateTest", "TiffTag306DateTimeTest"]
    suite = SuiteABC.from_target(target, required_tests, skipped_tests)
    suite_status = suite.compute_status()
    assert suite_status == SuiteStatus.AMBER


def test_that_a_suite_will_consider_required_tests_when_failing(test_targets):
    target = test_targets["bad"]
    required_tests = ["FileExtensionTest"]
    skipped_tests = ["LibTiffInfoTest", "GrepDateTest", "TiffTag306DateTimeTest"]
    suite = SuiteABC.from_target(target, required_tests, skipped_tests)
    suite_status = suite.compute_status()
    assert suite_status == SuiteStatus.RED


def test_that_a_suite_will_consider_required_tests_when_passing(test_targets):
    target = test_targets["good"]
    required_tests = ["Md5ChecksumTest"]
    suite = SuiteABC.from_target(target, required_tests)
    suite_status = suite.compute_status()
    assert suite_status == SuiteStatus.GREEN


def test_that_status_is_computed_if_not_already_assigned(test_targets):
    with patch.object(
        SuiteABC, "compute_status", return_value=SuiteStatus.GREEN
    ) as patch_compute_status:
        target = test_targets["good"]
        required_tests = ["Md5ChecksumTest"]
        suite = SuiteABC.from_target(target, required_tests)
        suite._status = SuiteStatus.NONE
        suite_status = suite.get_status()
        assert suite_status == SuiteStatus.GREEN
        patch_compute_status.assert_called_once()



import json

import pytest
from fs import open_fs

from dcqc.parsers import CsvParser
from dcqc.reports import JsonReport
from dcqc.tests.base_test import BaseTest
from dcqc.utils import open_parent_fs

PARENT_FOLDER_URL = "syn://syn50696607"


@pytest.fixture
def acceptance_test_folder_url(run_id):
    """Create a run-specific subfolder under parent folder.

    This is done to avoid clashes between concurrent tests.
    """
    fs = open_fs(PARENT_FOLDER_URL)
    fs.makedir(run_id)
    yield f"{PARENT_FOLDER_URL}/{run_id}"
    fs.removetree(run_id)


@pytest.mark.slow
def test_json_report_generation(get_data, acceptance_test_folder_url):
    # GIVEN a list of external tests to skip (to remain self-contained)
    all_tests = BaseTest.list_subclasses()
    skipped_tests = [test.__name__ for test in all_tests if test.is_external_test]

    # AND a subset of internal tests to be required (to verify suite status behavior)
    required_tests = ["Md5ChecksumTest"]

    # AND a CSV file of TXT and TIFF files
    csv_path = get_data("files.csv")

    # AND a remote destination for the JSON report
    report_url = f"{acceptance_test_folder_url}/report.json"

    # WHEN the CSV file is parsed to generate the relevant QC suites
    parser = CsvParser(csv_path)
    suites = parser.create_suites(required_tests, skipped_tests)

    # AND those suites are used to generate a JSON report
    report = JsonReport()
    report.save(suites, report_url, overwrite=True)

    # THEN the file exists
    fs, basename = open_parent_fs(report_url)
    assert fs.exists(basename)

    # AND the file can be loaded by the `json` module
    with fs.open(basename) as infile:
        contents = json.load(infile)
    assert contents



from dcqc.target import SingleTarget
from dcqc.tests.base_test import ExternalBaseTest, Process


class OmeXmlSchemaTest(ExternalBaseTest):
    tier = 2
    target: SingleTarget

    def generate_process(self) -> Process:
        path = self.target.file.stage()
        command_args = [
            "/opt/bftools/xmlvalid",
            path,
        ]
        process = Process(
            container="quay.io/sagebionetworks/bftools:latest",
            command_args=command_args,
        )
        return process



import pytest

from dcqc.reports import JsonReport


def test_for_error_when_creating_report_if_file_already_exists(get_data, test_files):
    existing_url = get_data("test.txt").as_posix()
    file = test_files["good"]
    report = JsonReport()
    with pytest.raises(FileExistsError):
        report.save(file, existing_url)


def test_that_a_single_object_can_be_reported_on(test_files):
    file = test_files["remote"]
    report_url = "mem://subdir/report.json"
    report = JsonReport()
    report.save(file, report_url, overwrite=True)


def test_for_an_error_when_saving_to_a_file_with_a_parent_being_a_file(test_files):
    file = test_files["good"]
    # Intentionally using the input file in the output path to trigger an error
    subdir_url = file.local_path.as_posix()
    report_url = f"{subdir_url}/report.json"

    report = JsonReport()
    with pytest.raises(NotADirectoryError):
        report.save(file, report_url)



from collections import defaultdict
from csv import DictWriter
from dataclasses import dataclass
from pathlib import Path
from typing import List

from dcqc.parsers import CsvParser
from dcqc.suites.suite_abc import SuiteABC


@dataclass
class CsvUpdater:
    input_path: Path
    output_path: Path
    parser: CsvParser

    def __init__(self, input_path: Path, output_path: Path):
        self.output_path = output_path
        self.input_path = input_path

    def update(self, suites: List[SuiteABC]):
        suite_dict = defaultdict(list)
        # {url: [list_of_statuses]} data structure to allow for multi-file targets
        # TODO add support for suites with multiple files in them (multi)
        for suite in suites:
            url = suite.target.files[0].url
            status = suite.get_status()
            suite_dict[url].append(status.value)
        # Evaluate dcqc_status for each url
        collapsed_dict = {}
        for url, statuses in suite_dict.items():
            if "RED" in statuses:
                collapsed_dict[url] = "RED"
            elif "AMBER" in statuses:
                collapsed_dict[url] = "AMBER"
            elif "GREEN" in statuses:
                collapsed_dict[url] = "GREEN"
            else:
                collapsed_dict[url] = "NONE"
        # Create CSV data structure
        row_list = []
        parser = CsvParser(self.input_path)
        for _, csv_data in parser.list_rows():
            csv_data["dcqc_status"] = collapsed_dict[csv_data["url"]]
            row_list.append(csv_data)

        if row_list:
            keys = row_list[0].keys()
            # Export updated CSV
            self.output_path.parent.mkdir(parents=True, exist_ok=True)
            with open(
                str(self.output_path), "w+", newline="", encoding="utf-8"
            ) as output_file:
                dict_writer = DictWriter(output_file, keys)
                dict_writer.writeheader()
                dict_writer.writerows(row_list)
        else:
            raise ValueError("No rows found in input CSV")



import json
from collections.abc import Iterable, Mapping
from pathlib import Path
from typing import Any, Optional, overload

from fs.base import FS
from fs.errors import ResourceNotFound

from dcqc.mixins import SerializableMixin, SerializedObject
from dcqc.utils import open_parent_fs


# TODO: Refactor instance methods to class methods
class JsonReport:
    paths_relative_to: Optional[Path]

    def __init__(self, paths_relative_to: Optional[Path] = None) -> None:
        self.paths_relative_to = paths_relative_to
        self._url: Optional[str] = None
        self._fs: Optional[FS] = None
        self._fs_path: Optional[str] = None

    # TODO: Move towards an FS mixin for these functions
    def _init_fs(self, url) -> tuple[FS, str]:
        self._url = url
        self._fs, self._fs_path = open_parent_fs(url)
        return self._fs, self._fs_path

    def _create_parent_directories(self, url: str):
        scheme, separator, resource = url.rpartition("://")
        parent_resource, _, _ = resource.rpartition("/")
        parent_url = f"{scheme}{separator}{parent_resource}"
        fs, fs_path = self._init_fs(parent_url)
        try:
            info = fs.getinfo(fs_path)
        except ResourceNotFound:
            fs.makedirs(fs_path, recreate=True)
            info = fs.getinfo(fs_path)
        if not info.is_dir:
            message = f"Parent URL ({url}) does not refer to a directory."
            raise NotADirectoryError(message)

    def to_file(self, obj: Any, url: str, overwrite: bool):
        fs, fs_path = self._init_fs(url)
        self._create_parent_directories(url)
        if fs.exists(fs_path) and not overwrite:
            message = f"URL ({url}) already exists. Enable `overwrite` to ignore."
            raise FileExistsError(message)
        # TODO: Implement custom serializer that handles Paths
        #       (e.g., relativize them based on output JSON path)
        with fs.open(fs_path, "w") as outfile:
            json.dump(obj, outfile, indent=2)
            outfile.write("\n")

    def _generate_single(self, item: SerializableMixin) -> SerializedObject:
        item.serialize_paths_relative_to(self.paths_relative_to)
        report = item.to_dict()
        return report

    # The overloads are necessary to convey the relationship between
    # the inputs and outputs: single to single, and many to many.
    @overload
    def generate(self, items: SerializableMixin) -> SerializedObject:
        """"""

    @overload
    def generate(self, items: Iterable[SerializableMixin]) -> list[SerializedObject]:
        """"""

    def generate(self, items):
        if isinstance(items, Iterable):
            report = [self._generate_single(item) for item in items]
        else:
            # In this else branch, `items` is actually a single item
            report = self._generate_single(items)
        return report

    @overload
    def save(
        self, items: SerializableMixin, url: str, overwrite: bool = False
    ) -> SerializedObject:
        """"""

    @overload
    def save(
        self, items: Iterable[SerializableMixin], url: str, overwrite: bool = False
    ) -> list[SerializedObject]:
        """"""

    def save(self, items, url: str, overwrite: bool = False):
        report = self.generate(items)
        self.to_file(report, url, overwrite)
        return report

    def save_many(
        self,
        named_items: Mapping[str, SerializableMixin],
        parent_url: str,
        overwrite: bool = False,
    ) -> dict[str, SerializedObject]:
        reports = dict()
        for name, item in named_items.items():
            report = self.generate(item)
            reports[name] = report
            report_url = f"{parent_url}/{name}"
            self.to_file(report, report_url, overwrite)
        return reports



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from dataclasses import fields
from itertools import chain
from pathlib import Path, PurePath
from typing import Any, ClassVar, Generic, Optional, Type, TypeVar, cast

SerializedObject = dict[str, Any]

T = TypeVar("T", bound="SerializableMixin")

U = TypeVar("U", bound=Type[object])


class SerializableMixin(ABC):
    # Used to serialize properties in addition to dataclass attributes
    _serialized_properties: ClassVar[list[str]]
    _serialized_properties = list()

    @classmethod
    def from_dict_prepare(cls, dictionary: SerializedObject) -> SerializedObject:
        """Validate and prepare dictionary for deserialization."""
        type_ = dictionary.pop("type")
        if type_ != cls.__name__:  # pragma: no cover
            message = f"Type ({type_}) does not match the class ({cls.__name__})."
            raise ValueError(message)
        return dictionary

    # TODO: Move this logic to JsonReport as a JSONEncoder subclass
    def serialize_paths_relative_to(self, location: Optional[Path]):
        if location is not None and not (location.exists() and location.is_dir()):
            message = f"Location ({location}) is not an existing directory."
            raise ValueError(message)
        self._serialize_paths_relative_to = location

    def serialize_path(self, path: PurePath) -> str:
        # This is useful for portability between steps in Nextflow
        if getattr(self, "_serialize_paths_relative_to", None):
            path_str = os.path.relpath(path, self._serialize_paths_relative_to)
        else:
            path_str = str(path)
        return path_str

    def serialize_value(self, value: Any) -> Any:
        """Ensure that all values are JSON-serializable.

        Args:
            value: Any value.

        Returns:
            An equivalent JSON-serializable value.
        """
        result: Any
        if isinstance(value, (str, bytes)):
            result = value
        elif isinstance(value, PurePath):
            result = self.serialize_path(value)
        elif isinstance(value, SerializableMixin):
            result = value.to_dict()
        elif isinstance(value, Mapping):
            result = {key: self.serialize_value(val) for key, val in value.items()}
        elif isinstance(value, Iterable):
            result = [self.serialize_value(item) for item in value]
        else:
            result = value
        return result

    def dict_factory(self, iterable: list[tuple[str, Any]]) -> dict[str, Any]:
        """Generate dictionary from dataclass.

        Unlike the built-in version, this function will
        handle Path objects. This assumes that the OS
        will not change between the serialization and
        deserialization steps.

        Args:
            iterable: List of attribute name-value pairs.

        Returns:
            Dictionary of JSON-serializable attributes.
        """
        kwargs = {}
        for key, value in iterable:
            kwargs[key] = self.serialize_value(value)
        return dict(**kwargs)

    def to_dict(self) -> SerializedObject:
        """Serialize the file to a dictionary.

        Inspired by dataclasses.asdict().

        Returns:
            A file serialized as a dictionary.
        """
        result = []
        official_fields = [field.name for field in fields(self)]
        serialized_properties = getattr(self, "_serialized_properties", [])
        for field in chain(official_fields, serialized_properties):
            # TODO: Code smell indicating that some restructuring is in order
            try:
                value_raw = getattr(self, field)
            except Exception:
                value_raw = None
            value = self.serialize_value(value_raw)
            result.append((field, value))
        return self.dict_factory(result)

    # TODO: Use template method to handle `_serialized_properties`
    #       as well as `deepcopy()`
    @classmethod
    @abstractmethod
    def from_dict(cls, dictionary: SerializedObject) -> SerializableMixin:
        """Deserialize a dictionary into a SerializableMixin object.

        Args:
            dictionary: A serialized object.

        Returns:
            The reconstructed object.
        """

    def copy(self: T) -> T:
        """Create a copy of a serializable object.

        Returns:
            A copied object.
        """
        dictionary = self.to_dict()
        copy = self.from_dict(dictionary)
        # Required to prevent this mypy error:
        # Incompatible return value type (got "SerializableMixin", expected "T")
        copy = cast(T, copy)
        return copy


class SubclassRegistryMixin(ABC, Generic[U]):
    """Mixin for tracking subclasses."""

    @classmethod
    @abstractmethod
    def get_base_class(cls):
        """Retrieve base class."""

    @classmethod
    def list_subclasses(cls) -> tuple[Type[U], ...]:
        """List all subclasses."""
        subclasses = cls.__subclasses__()
        subsubclasses_list = [subcls.list_subclasses() for subcls in subclasses]
        subclasses_chain = chain(subclasses, *subsubclasses_list)
        all_subclasses = tuple(dict.fromkeys(subclasses_chain))
        return all_subclasses

    @classmethod
    def get_subclass_by_name(cls, name: str) -> Type[U]:
        """Retrieve a subclass by name."""
        subclasses = cls.get_base_class().list_subclasses()
        registry = {subcls.__name__: subcls for subcls in subclasses}
        if name not in registry:
            options = list(registry)
            message = f"Subclass ({name}) not available ({options})."
            raise ValueError(message)
        return registry[name]



from collections.abc import Generator

import pytest

from dcqc.file import File
from dcqc.mixins import SerializableMixin
from dcqc.parsers import CsvParser, JsonParser
from dcqc.suites.suite_abc import SuiteABC
from dcqc.target import SingleTarget
from dcqc.tests.base_test import BaseTest


def test_that_parsing_a_csv_file_yields_suites(get_data):
    csv_path = get_data("files.csv")
    csv_parser = CsvParser(csv_path)
    result = csv_parser.create_suites()
    assert isinstance(result, Generator)
    result = list(result)
    assert len(result) > 1
    assert all(isinstance(x, SuiteABC) for x in result)


def test_that_parsing_a_csv_file_stages_remote_files(get_data, test_files, mocker):
    csv_path = get_data("files.csv")
    csv_parser = CsvParser(csv_path, stage_files=True)
    mock = mocker.patch.object(csv_parser, "_row_to_file")
    mock.return_value = test_files["remote"]
    files = csv_parser.create_files()
    assert all(file.local_path is not None for _, file in files)


def test_that_parsing_a_json_file_must_match_listed_type(get_data):
    json_path = get_data("file.json")
    assert JsonParser.parse_object(json_path, File)
    with pytest.raises(ValueError):
        JsonParser.parse_object(json_path, SingleTarget)


def test_for_an_error_when_parsing_an_unrecognized_type():
    with pytest.raises(ValueError):
        JsonParser.get_class("foobar")


def test_for_an_error_when_parsing_a_dictionary_without_a_type():
    dictionary = {"foo": "bar"}
    with pytest.raises(ValueError):
        JsonParser.from_dict(dictionary)


def test_for_an_error_when_parsing_a_list_of_objects_with_parse_object(get_data):
    json_path = get_data("tests.json")
    with pytest.raises(ValueError):
        JsonParser.parse_object(json_path, SerializableMixin)


def test_for_an_error_when_parsing_a_single_object_with_parse_objects(get_data):
    json_path = get_data("target.json")
    with pytest.raises(ValueError):
        JsonParser.parse_objects(json_path, SerializableMixin)


def test_that_json_parser_can_parse_multiple_objects(get_data):
    json_path = get_data("tests.json")
    result = JsonParser.parse_objects(json_path, BaseTest)
    assert len(result) > 0
    assert isinstance(result[0], BaseTest)



import pytest

from dcqc.target import PairedTarget, SingleTarget


def test_that_a_target_can_be_saved_and_restored_without_changing(test_files):
    test_file = test_files["good"]
    target_1 = SingleTarget(test_file)
    target_1_dict = target_1.to_dict()
    target_2 = SingleTarget.from_dict(target_1_dict)
    target_2_dict = target_2.to_dict()
    assert target_1 == target_2
    assert target_1_dict == target_2_dict


def test_for_an_error_when_restoring_a_target_with_a_discordant_type(test_files):
    test_file = test_files["good"]
    target_1 = SingleTarget(test_file)
    target_1_dict = target_1.to_dict()
    target_1_dict["type"] = "UnexpectedQcTarget"
    with pytest.raises(ValueError):
        SingleTarget.from_dict(target_1_dict)


def test_for_an_error_when_creating_single_file_target_with_two_files(test_files):
    test_file = test_files["good"]
    with pytest.raises(ValueError):
        SingleTarget([test_file, test_file])


def test_that_paired_file_target_can_be_created_with_two_files(test_files):
    test_file = test_files["good"]
    target = PairedTarget([test_file, test_file])
    assert len(target.files) == 2


def test_for_an_error_when_creating_paired_file_target_with_one_file(test_files):
    test_file = test_files["good"]
    with pytest.raises(ValueError):
        PairedTarget(test_file)


def test_for_an_error_when_creating_paired_file_target_with_one_file_in_list(
    test_files,
):
    test_file = test_files["good"]
    with pytest.raises(ValueError):
        PairedTarget([test_file])



from __future__ import annotations

from abc import ABC
from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from typing import Optional

from dcqc.file import File, FileType
from dcqc.mixins import SerializableMixin, SerializedObject, SubclassRegistryMixin


# TODO: Eventually, there might be target-specific metadata
@dataclass
class BaseTarget(SerializableMixin, SubclassRegistryMixin, ABC):
    """Base class for targets with one or more files.

    Attributes:
        files: List of files objects.
        id: A unique identifier for the target. Defaults to None.
        type: The target type/subclass.
    """

    file_or_files: InitVar[File | list[File]]
    id: Optional[str] = None
    files: list[File] = field(init=False)
    type: str = field(init=False)

    def __post_init__(self, file_or_files: File | list[File]):
        """Ensure list of files and fill in Target type."""
        self.type = self.__class__.__name__
        if isinstance(file_or_files, File):
            self.files = [file_or_files]
        else:
            self.files = file_or_files

    def stage(
        self,
        destination: Optional[Path] = None,
        overwrite: bool = False,
    ) -> list[Path]:
        """Create local copy of local or remote file.

        A destination is not required for remote files; it
        defaults to a temporary directory.
        Local files aren't moved if a destination is omitted.

        Args:
            destination: File or folder where to store the file.
                Defaults to None.
            overwrite: Whether to ignore existing file at the
                target destination. Defaults to False.

        Raises:
            ValueError: If the parent directory of the
                destination does not exist.
            FileExistsError: If the destination file already
                exists and ``overwrite`` was not enabled.

        Returns:
            The path of the local copy.
        """
        paths = list()
        for file in self.files:
            path = file.stage(destination, overwrite)
            paths.append(path)
        return paths

    @classmethod
    def from_dict(cls, dictionary: SerializedObject) -> BaseTarget:
        """Deserialize a dictionary into a target.

        Args:
            dictionary: A serialized target object.

        Returns:
            The reconstructed target object.
        """
        target_cls_name = dictionary["type"]
        target_cls = BaseTarget.get_subclass_by_name(target_cls_name)
        dictionary = deepcopy(dictionary)
        dictionary = target_cls.from_dict_prepare(dictionary)
        files = [File.from_dict(d) for d in dictionary["files"]]
        id = dictionary["id"]
        target = target_cls(*files, id=id)
        return target

    @classmethod
    def get_base_class(cls):
        """Retrieve base class."""
        return BaseTarget


@dataclass(init=False)
class SingleTarget(BaseTarget):
    """Single-file target."""

    def __post_init__(self, file_or_files: File | list[File]):
        """Run validation checks after initialization."""
        super().__post_init__(file_or_files)
        self.ensure_single_file()

    # While this function makes sense as a pydantic validator,
    # we can into strange issues with the following test after
    # switching to @pydantic.dataclasses.dataclass:
    # test_that_paths_are_unchanged_when_not_using_serialize_paths_relative_to
    def ensure_single_file(self):
        """Ensure that target is only initialized with a single file."""
        if len(self.files) != 1:
            raise ValueError("SingleTarget is restricted to single files")

    @property
    def file(self):
        """Single file."""
        return self.files[0]

    def get_file_type(self) -> FileType:
        """Retrieve the file type for the target.

        Returns:
            The file type object.
        """
        return self.file.get_file_type()


@dataclass(init=False)
class PairedTarget(BaseTarget):
    """Paired (two-file) target."""

    def __post_init__(self, file_or_files: File | list[File]):
        """Run validation checks after initialization."""
        super().__post_init__(file_or_files)
        self.ensure_two_files()

    def ensure_two_files(self):
        """Ensure that target is only initialized with two files."""
        if len(self.files) != 2:
            raise ValueError("PairedTarget is restricted to two files")



import hashlib
from pathlib import Path

from dcqc.target import SingleTarget
from dcqc.tests.base_test import InternalBaseTest, TestStatus


class Md5ChecksumTest(InternalBaseTest):
    tier = 1
    target: SingleTarget

    def compute_status(self) -> TestStatus:
        path = self.target.file.stage()
        expected_md5 = self.target.file.get_metadata("md5_checksum")
        actual_md5 = self._compute_md5_checksum(path)
        if expected_md5 == actual_md5:
            status = TestStatus.PASS
        else:
            status = TestStatus.FAIL
        return status

    def _compute_md5_checksum(self, path: Path) -> str:
        hash_md5 = hashlib.md5()
        with path.open("rb") as infile:
            for chunk in iter(lambda: infile.read(4096), b""):
                hash_md5.update(chunk)
        actual_md5 = hash_md5.hexdigest()
        return actual_md5



import shutil
from subprocess import check_output
from typing import Any

import pytest
from click.testing import Result
from typer.testing import CliRunner

import dcqc
from dcqc.main import app
from dcqc.parsers import JsonParser
from dcqc.suites.suite_abc import SuiteABC


def run_command(arguments: list[Any]):
    runner = CliRunner()
    str_arguments = [str(arg) for arg in arguments]
    result = runner.invoke(app, str_arguments)
    return result


def check_command_result(result: Result):
    if result.exit_code != 0:
        print(result.stdout)
        try:
            print(result.stderr)
        except ValueError:
            pass
    assert result.exit_code == 0


@pytest.mark.slow
def test_that_the_module_cli_behaves_the_same_as_the_plain_cli():
    module_cli = check_output(["python", "-m", "dcqc", "--help"])
    plain_cli = check_output(["dcqc", "--help"])
    assert module_cli == plain_cli


def test_that_the_package_version_can_be_printed():
    args = ["--version"]
    result = run_command(args)
    check_command_result(result)
    assert dcqc.__version__ in result.output


def test_create_targets(get_data, get_output):
    input_csv = get_data("small.csv")
    output_dir = get_output("create_targets")
    shutil.rmtree(output_dir, ignore_errors=True)

    assert not output_dir.exists()
    args = ["create-targets", input_csv, output_dir]
    result = run_command(args)
    check_command_result(result)
    assert len(list(output_dir.iterdir())) > 0


def test_create_tests(get_data, get_output):
    input_json = get_data("target.json")
    output_dir = get_output("create_tests")
    shutil.rmtree(output_dir, ignore_errors=True)

    assert not output_dir.exists()
    args = ["create-tests", "-r", "Md5ChecksumTest", input_json, output_dir]
    result = run_command(args)
    check_command_result(result)
    assert len(list(output_dir.iterdir())) > 0


def test_create_process(get_data, get_output):
    input_json = get_data("test.external.json")
    output_path = get_output("create_process") / "process.json"
    output_path.unlink(missing_ok=True)

    assert not output_path.exists()
    args = ["create-process", input_json, output_path]
    result = run_command(args)
    check_command_result(result)
    assert output_path.exists()


def test_compute_test(get_data, get_output):
    input_json = get_data("test.internal.json")
    output_path = get_output("compute_test") / "test.json"
    output_path.unlink(missing_ok=True)

    assert not output_path.exists()
    args = ["compute-test", input_json, output_path]
    result = run_command(args)
    check_command_result(result)
    assert output_path.exists()


def test_create_suite(get_data, get_output):
    input_json = get_data("test.computed.json")
    output_path = get_output("create_suite") / "suite.json"
    output_path.unlink(missing_ok=True)

    args = ["create-suite", output_path, input_json, input_json, input_json]
    result = run_command(args)
    check_command_result(result)
    assert output_path.exists()

    suite = JsonParser.parse_object(output_path, SuiteABC)
    assert len(suite.required_tests) > 0


def test_combine_suites(get_data, get_output):
    input_json = get_data("suite.json")
    output_path = get_output("combine_suites") / "suites.json"
    output_path.unlink(missing_ok=True)

    args = ["combine-suites", output_path, input_json, input_json, input_json]
    result = run_command(args)
    check_command_result(result)
    assert output_path.exists()


def test_list_tests():
    args = ["list-tests"]
    result = run_command(args)
    check_command_result(result)


def test_qc_file(get_data):
    tiff_path = get_data("circuit.tif")
    args = [
        "qc-file",
        "-t",
        "TIFF",
        "-m",
        '{"md5_checksum": "c7b08f6decb5e7572efbe6074926a843"}',
        tiff_path,
    ]
    result = run_command(args)
    check_command_result(result)


def test_update_csv(get_data, get_output):
    suites_path = get_data("suites.json")
    input_path = get_data("input.csv")
    output_path = get_output("update_csv") / "output.csv"
    output_path.unlink(missing_ok=True)

    args = [
        "update-csv",
        suites_path,
        input_path,
        output_path,
    ]
    result = run_command(args)
    check_command_result(result)
    assert output_path.exists()



import csv

import pytest

from dcqc.updaters import CsvUpdater


def get_dcqc_status_list_from_file(filename):
    with open(filename, "r") as file:
        reader = csv.DictReader(file)
        status_list = [row["dcqc_status"] for row in reader]
    return status_list


def test_that_csv_updater_updates_csv_as_expected_with_single_targets(
    get_data, mocked_suites_single_targets
):
    input_file = get_data("test_input.csv")
    output_file = get_data("test_output.csv")
    updater = CsvUpdater(input_file, output_file)
    updater.update(mocked_suites_single_targets)
    status_list = get_dcqc_status_list_from_file(output_file)
    assert status_list == ["GREEN", "RED", "AMBER", "NONE"]


def test_that_empty_input_manifest_raises_error(get_data, mocked_suites_single_targets):
    with pytest.raises(ValueError):
        empty_updater = CsvUpdater(
            get_data("empty_input.csv"), get_data("test_output.csv")
        )
        empty_updater.update(mocked_suites_single_targets)


# def test_that_csv_updater_updates_csv_as_expected_with_multi_targets(
#     get_data, mocked_suites_multi_targets
# ):
#     input_file = get_data("input.csv")
#     output_file = get_data("output.csv")
#     updater = CsvUpdater(input_file, output_file)
#     updater.update(mocked_suites_multi_targets)
#     status_list = get_dcqc_status_list_from_file(output_file)
#     assert status_list == ["GREEN", "RED", "AMBER"]



#!/usr/bin/env python3
# TODO move this to pytest fixtures

"""Generate test files using latest version of dcqc."""

import os
import sys
from pathlib import Path
from typing import Sequence

from dcqc import tests
from dcqc.file import File
from dcqc.mixins import SerializableMixin
from dcqc.parsers import JsonParser
from dcqc.reports import JsonReport
from dcqc.suites.suite_abc import SuiteABC
from dcqc.target import SingleTarget

# Shared values
data_dir = sys.path[0]
data_dir = os.path.relpath(data_dir)
report = JsonReport(paths_relative_to=Path.cwd())


# Shared functions
def export(obj: SerializableMixin | Sequence[SerializableMixin], filename: str):
    output_url = os.path.join(data_dir, filename)
    report.save(obj, output_url, overwrite=True)


# file.json
file_url = os.path.join(data_dir, "test.txt")
metadata = {"file_type": "TIFF", "md5_checksum": "14758f1afd44c09b7992073ccf00b43d"}
file = File(file_url, metadata)
export(file, "file.json")

# target.json
target = SingleTarget(file, id="001")
export(target, "target.json")

# test.internal.json
internal_test = tests.Md5ChecksumTest(target)
export(internal_test, "test.internal.json")

# test.external.json
external_test = tests.LibTiffInfoTest(target)
export(external_test, "test.external.json")

# test.computed.json
computed_test = tests.Md5ChecksumTest(target)
computed_test.get_status()
export(computed_test, "test.computed.json")

# tests.json
test_list = [internal_test, external_test, computed_test]
export(test_list, "tests.json")

# suite.json
suite_tests = [internal_test, external_test]
required_tests = ["Md5ChecksumTest"]
skipped_tests = ["LibTiffInfoTest"]
suite = SuiteABC.from_tests(suite_tests, required_tests, skipped_tests)
export(suite, "suite.json")

# suites.json
input_jsons = [
    Path(file_path)
    for file_path in [
        "tests/data/suites_files/suites_1.json",
        "tests/data/suites_files/suites_2.json",
        "tests/data/suites_files/suites_3.json",
    ]
]
suites = [JsonParser.parse_object(json_, SuiteABC) for json_ in input_jsons]
export(suites, "suites.json")



import gzip
from pathlib import Path
from typing import TextIO

from dcqc.target import PairedTarget
from dcqc.tests.base_test import InternalBaseTest, TestStatus


class PairedFastqParityTest(InternalBaseTest):
    tier = 2
    target: PairedTarget

    def compute_status(self) -> TestStatus:
        """Compute test status."""
        counts = list()
        for file in self.target.files:
            path = file.stage()
            try:
                count = self._count_fastq_lines(path)
            except Exception:
                return TestStatus.FAIL
            counts.append(count)

        # Check that there counts are all the same (i.e., equal)
        if len(set(counts)) <= 1:
            status = TestStatus.PASS
        else:
            status = TestStatus.FAIL
        return status

    def _count_fastq_lines(self, path: Path) -> int:
        """Count the number of lines in a FASTQ file.

        Args:
            path: Path to the FASTQ file.

        Returns:
            Number of lines in the given FASTQ file.
        """
        # Source: https://stackoverflow.com/a/1019572/21077945
        with self._open_fastq(path) as fastq:
            num_lines = sum(1 for _ in fastq)
        return num_lines

    def _open_fastq(self, path: Path) -> TextIO:
        """Open a FASTQ file regardless of compression.

        Args:
            path: Path to the FASTQ file.

        Returns:
            Opened FASTQ file (in text mode).
        """
        # TODO: This logic should ideally live in the File class, and a
        #       test should confirm the integrity of compressed files
        if path.name.endswith(".gz"):
            return gzip.open(path, "rt")
        else:
            return path.open("rt")
